import os
from os.path import basename, isfile, join
import sys

import nibabel as nib
from nilearn.image import reorder_img
import numpy as np
import onnxruntime as ort

model_servers = ["https://github.com/StanleyWangTW/tigersyn/releases/download/v0.0.2-alpha/mprage_syntheseg_v001_unet.onnx"]
# determine if application is a script file or frozen exe
if getattr(sys, 'frozen', False):
    application_path = os.path.dirname(sys.executable)
elif __file__:
    application_path = os.path.dirname(os.path.abspath(__file__))

model_path = join(application_path, 'models')
print(model_path)
os.makedirs(model_path, exist_ok=True)

label_all = dict()
label_all['syntheseg'] = (
    2, 3, 4, 5, 7, 8, 10, 11, 12, 13, 14, 15, 16, 17, 18, 26,
    28, 41, 42, 43, 44, 46, 47, 49, 50, 51, 52, 53, 54, 58, 60
)

def read_nib(input_nib):
    """nib to 5D numpy"""
    return input_nib.get_fdata()[None, None, ...]

def save_nib(data_nib, ftemplate, postfix):
    output_file = ftemplate.replace('@@@@', postfix)
    nib.save(data_nib, output_file)
    print('Writing output file: ', output_file)
    return output_file

def read_file(model_ff, input_file):
    """load nib and reorder"""
    input_nib = nib.load(input_file)
    vol_nib = reorder_img(input_nib, resample="continuous")
    return vol_nib


def download(url, file_name):
    import urllib.request
    import certifi
    import shutil
    import ssl
    context = ssl.create_default_context(cafile=certifi.where())
    with urllib.request.urlopen(url,
                                context=context) as response, open(file_name, 'wb') as out_file:
        shutil.copyfileobj(response, out_file)

def get_model(f):
    if isfile(f):
        return f
    
    if '.onnx' in f:
        fn = f
    else:
        fn = f + '.onnx'

    model_file = join(model_path, fn)

    if not os.path.exists(model_file):
        
        for server in model_servers:
            try:
                print(f'Downloading model files....')
                model_url = server
                print(model_url, model_file)
                download(model_url, model_file)
                download_ok = True
                print('Download finished...')
                break
            except:
                download_ok = False

        if not download_ok:
            raise ValueError('Server error. Please check the model name or internet connection.')
                
    return model_file


def get_mode(model_ff):
    seg_mode, version, model_str = basename(model_ff).split('_')[1:4]  # aseg43, bet
    #print(seg_mode, version , model_str)

    return seg_mode, version, model_str

def normalize(data):
    return (data - data.min()) / (data.max() - data.min())

def predict(model, data, GPU):
    """read array-like data, then segmentation"""
    if GPU and (ort.get_device() == "GPU"):
        ort_sess = ort.InferenceSession(model, providers=['CUDAExecutionProvider'])
    else:
        ort_sess = ort.InferenceSession(model, providers=["CPUExecutionProvider"])

    data_type = 'float32'
    return ort_sess.run(None, {'input': data.astype(data_type)})[0]

def run(model_ff, input_nib, GPU):
    """segment nib, return mask_nib"""
    seg_mode, _ , model_str = get_mode(model_ff)

    data = read_nib(input_nib)
    # data = normalize(data)

    logits = predict(model_ff, data, GPU)[0, ...]
    mask_pred = np.argmax(logits, axis=0)

    if seg_mode in ['syntheseg']:
        labels = label_all[seg_mode]
        mask_pred_relabel = mask_pred * 0
        for ii in range(len(labels)):
            mask_pred_relabel[mask_pred == (ii + 1)] = labels[ii]
            #print((ii+1), labels[ii])
        mask_pred = mask_pred_relabel
    
    mask_pred = mask_pred.astype(int)
    output_nib = nib.Nifti1Image(mask_pred, input_nib.affine, input_nib.header)
    
    return output_nib


import os
from os.path import basename, join
import glob
import time

import nibabel as nib
from nilearn.image import resample_to_img
import numpy as np

from tigersyn import syn_tool

model_syn = r"mprage_syntheseg_v001_unet.onnx"

def produce_mask(model, f, GPU):
    model_ff = syn_tool.get_model(model)
    print(model_ff)
    input_nib = nib.load(f)
    input_nib_resp = syn_tool.read_file(model_ff, f)

    mask_nib_resp = syn_tool.run(
        model_ff, input_nib_resp,  GPU=GPU)
    
    mask_nib = resample_to_img(mask_nib_resp, input_nib, interpolation="nearest")

    output =  mask_nib.get_fdata()

    if np.max(output) <=255:
        dtype = np.uint8
    else:
        dtype = np.int16

    output = output.astype(dtype)
    
    output_nib = nib.Nifti1Image(output, input_nib.affine, input_nib.header)
    output_nib.header.set_data_dtype(dtype)

    return output_nib


def run(argstring, input, output=None, model=None):
    syn = "s" in argstring
    gpu = "g" in argstring
    get_z = 'z' in argstring

    if not isinstance(input, list):
        input = [input]

    input_file_list = input
    if os.path.isdir(input[0]):
        input_file_list = glob.glob(join(input[0], '*.nii'))
        input_file_list += glob.glob(join(input[0], '*.nii.gz'))

    output_dir = output

    print('Total nii files:', len(input_file_list))
    count = 0
    for f in input_file_list:
        count += 1
        t = time.time()

        f_output_dir = output_dir
        if f_output_dir is None:
            f_output_dir = os.path.dirname(os.path.abspath(f))
        else:
            os.makedirs(f_output_dir, exist_ok=True)
        
        print(f'{count} Processing :', os.path.basename(f))

        ftemplate = basename(f).replace('.nii', f'_@@@@.nii')
        if get_z and '.gz' not in ftemplate:
            ftemplate += '.gz'
        ftemplate = join(f_output_dir, ftemplate)
        
        if syn:
            aseg_nib = produce_mask(model_syn, f, GPU=gpu)
            fn = syn_tool.save_nib(aseg_nib, ftemplate, 'syn')

        print('Processing time: %d seconds' %  (time.time() - t))

if __name__ == "__main__":
    run("sz", r"T1w.nii.gz", r"outputs")


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

VESION = '0.0.2a1'
DESCRIPTION = 'Processing MRI images based on deep-learning'
CLASSIFIERS = [
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3.9',
    'License :: OSI Approved :: MIT License',
    "Operating System :: Microsoft :: Windows"
]

setup(
    name='tigersyn',
    url='https://github.com/StanleyWangTW/tigersyn',

    version=VESION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='X. S. Wang',
    author_email='',
    License='MIT',

    packages=find_packages(),
    classifiers=CLASSIFIERS,
    python_requires='>=3.9',
    install_requires=[
        'numpy',
        'nibabel',
        'nilearn',
        'onnxruntime'
    ]
)